@@ -1729,23 +1729,16 @@ static void get_info(struct SYMBOL *s,
 			}
 		}
 
-		{
-			int i;
-
-			for (i = 'A' - 'A'; i <= 'Z' - 'A'; i++) {
-				if (i == 'T' - 'A' || i == 'X' - 'A')
-					continue;
-				if (!(cfmt.fields[0] & (1 << i)))
-					info[i] = 0;
-			}
-/*fixme:do it also for 'a'..'z'?*/
-		}
 		nbar = nbar_rep = cfmt.measurefirst;	/* measure numbering */
 		over_voice = -1;
 		over_time = -1;
 		over_bar = 0;
 		reset_gen();
 
+		/* remove unwanted fields */
+		if (!(cfmt.fields[0] & (1 << ('Q' - 'A'))))
+			info['Q' - 'A'] = 0;
+
 		/* switch to the 1st voice */
 		curvoice = &voice_tb[parsys->top_voice];
 
@@ -1855,8 +1848,10 @@ static void get_info(struct SYMBOL *s,
 			break;
 		goto addinfo;
 	case 'X':
-		if (!epsf)
-			write_buffer();	/* flush stuff left from %% lines */
+		if (!epsf) {
+			buffer_eob();	/* flush stuff left from %% lines */
+			write_buffer();
+		}
 		dfmt = cfmt;		/* save format and info */
 		memcpy(&info_glob, &info, sizeof info_glob);
 		memcpy(&deco_tune, &deco_glob, sizeof deco_tune);
@@ -2554,9 +2549,12 @@ static void get_key(struct SYMBOL *s)
 				curvoice->key.sf = 0;
 			break;
 		}
-		if (curvoice->ckey.sf == s->as.u.key.sf	/* if same key */
-		 && curvoice->ckey.nacc == s->as.u.key.nacc
-		 && s->as.next->type != ABC_T_CLEF)	/* but not explicit clef */
+		if ((s->as.next == 0
+		  || s->as.next->type != ABC_T_CLEF)	/* if not explicit clef */
+		 && curvoice->ckey.sf == s->as.u.key.sf	/* and same key */
+		 && curvoice->ckey.nacc == 0
+		 && s->as.u.key.nacc == 0
+		 && curvoice->ckey.empty == s->as.u.key.empty)
 			break;				/* ignore */
 
 		if (!curvoice->ckey.empty)
@@ -2934,8 +2932,10 @@ static struct abcsym *process_pscomment(struct abcsym *as)
 				use = 'g';
 			}
 			p = as->text + 2 + 7;
-			while (*p != '\n')
+			while (*p != '\0' && *p != '\n')
 				p++;
+			if (*p == '\0')
+				return as;		/* empty */
 			p++;
 			ps_def((struct SYMBOL *) as, p, use);
 			return as;
@@ -2944,8 +2944,10 @@ static struct abcsym *process_pscomment(struct abcsym *as)
 			int job;
 
 			gen_ly(1);
-			if (epsf && as->state == ABC_S_GLOBAL)
-				return as;
+			if (as->state == ABC_S_GLOBAL) {
+				if (epsf || in_fname == 0)
+					return as;
+			}
 			p = as->text + 2 + 9;
 			while (*p == ' ' || *p == '\t')
 				p++;
@@ -3011,7 +3013,7 @@ static struct abcsym *process_pscomment(struct abcsym *as)
 			}
 			PUT0("\001");	/* include file (must be the first after eob) */
 			bskip(y2 - y1);
-			PUT3("%.2f %.2f%%%s\n", x1, y1, fn);
+			PUT3("%.2f %.2f%%%s\n", -x1, -y1, fn);
 			buffer_eob();
 			return as;
 		}
@@ -3085,7 +3087,7 @@ static struct abcsym *process_pscomment(struct abcsym *as)
 		break;
 	case 'n':
 		if (strcmp(w, "newpage") == 0) {
-			if (epsf)
+			if (epsf || in_fname == 0)
 				return as;
 			generate();
 			buffer_eob();
@@ -3175,6 +3177,10 @@ irepeat:
 			float h2, len, lwidth;
 
 			gen_ly(0);
+			if (as->state == ABC_S_GLOBAL) {
+				if (epsf || in_fname == 0)
+					return as;
+			}
 			lwidth = (cfmt.landscape ? cfmt.pageheight : cfmt.pagewidth)
 				- cfmt.leftmargin - cfmt.rightmargin;
 			h1 = h2 = len = 0;
@@ -3265,8 +3271,10 @@ irepeat:
 			int job;
 
 			gen_ly(1);
-			if (epsf && as->state == ABC_S_GLOBAL)
-				return as;
+			if (as->state == ABC_S_GLOBAL) {
+				if (epsf || in_fname == 0)
+					return as;
+			}
 			if (w[0] == 'c') {
 				job = T_CENTER;
 			} else {
@@ -3325,6 +3333,10 @@ irepeat:
 	case 'v':
 		if (strcmp(w, "vskip") == 0) {
 			gen_ly(0);
+			if (as->state == ABC_S_GLOBAL) {
+				if (epsf || in_fname == 0)
+					return as;
+			}
 			h1 = scan_u(p);
 			bskip(h1);
 			buffer_eob();
@@ -3352,6 +3364,8 @@ irepeat:
 			curvoice->transpose = cfmt.transpose;
 			s2 = curvoice->sym;
 			if (s2 == 0) {
+				memcpy(&curvoice->key, &curvoice->okey,
+					sizeof curvoice->key);
 				key_transpose(&curvoice->key);
 				memcpy(&curvoice->ckey, &curvoice->key,
 					sizeof curvoice->ckey);
@@ -3363,8 +3377,6 @@ irepeat:
 				if (s2->type == KEYSIG)
 					break;
 			if (s2 == 0 || s2->time != curvoice->time) {
-				memcpy(&s->as.u.key, &curvoice->key,
-							sizeof s->as.u.key);
 				s->as.type = ABC_T_INFO;
 				s->as.text[0] = 'K';
 				s->as.text[1] = '\0';
@@ -3372,6 +3384,8 @@ irepeat:
 				s->u = curvoice->ckey.sf;
 				s2 = s;
 			}
+			memcpy(&s2->as.u.key, &curvoice->okey,
+						sizeof s2->as.u.key);
 			key_transpose(&s2->as.u.key);
 			memcpy(&curvoice->ckey, &s->as.u.key,
 						sizeof curvoice->ckey);
