@@ -156,14 +156,23 @@ static void cap_str(char *p)
 {
 	while (*p != '\0') {
 #if 1
-/* pb with toupper - works with ASCII only */
+/* pb with toupper - works with ASCII and some latin characters only */
 		unsigned char c;
 
 		c = (unsigned char) *p;
-		if (c >= 'a' && c <= 'z')
-/*fixme: KO with utf-8*/
-//		 || (c >= 0xe0 && c <= 0xfe))
+		if (c >= 'a' && c <= 'z') {
 			*p = c & ~0x20;
+		} else if (c == 0xc3) {
+			p++;
+			c = *p;
+			if (c >= 0xa0 && c <= 0xbe && c != 0xb7)
+				*p = c & ~0x20;
+		} else if (c == 0xc4) {
+			p++;
+			c = *p;
+			if (c >= 0x81 && c <= 0xb7 && (c & 0x01))
+				(*p)--;
+		}
 #else
 		*p = toupper((unsigned char) *p);
 #endif
@@ -685,7 +694,6 @@ static void pg_para_output(int job)
 	bskip(y);
 	pango_layout_set_attributes(layout, NULL);
 	pg_str = g_string_truncate(pg_str, 0);
-	pango_attr_list_unref(attrs);
 }
 
 /* output of filled / justified text*/
@@ -720,7 +728,9 @@ static void pg_write_text(char *s, int job, float baseskip)
 	str_set_font(tex_buf);
 	if (pg_str->len)
 		pg_para_output(job);
+	pango_attr_list_unref(attrs);
 }
+
 /* check if pango is needed */
 static int is_latin(unsigned char *p)
 {
@@ -1046,6 +1056,10 @@ void write_text(char *cmd, char *s, int job)
 				p++;
 			if (*p != '\0')
 				*p++ = '\0';
+			if (*s == '\0') {
+				bskip(baseskip * 0.5);
+				buffer_eob();
+			}
 			bskip(baseskip);
 			switch (job) {
 			case A_LEFT:
@@ -1216,11 +1230,6 @@ static int put_wline(char *p,
 			p++;
 	}
 
-	/* on the left side, permit page break at empty lines or stanza start */
-	if (!right
-	 && (*p == '\0' || r != 0))
-		buffer_eob();
-
 	if (r != 0) {
 		sep = *r;
 		*r = '\0';
@@ -1295,6 +1304,9 @@ void put_words(struct SYMBOL *words)
 	/* output the text */
 	bskip(cfmt.wordsspace);
 	for (s = words; s != 0 || s2 != 0; ) {
+//fixme:should also permit page break on stanza start
+		if (s != 0 && s->as.text[2] == '\0')
+			buffer_eob();
 		bskip(cfmt.lineskipfac * cfmt.font_tb[WORDSFONT].size);
 		if (s != 0) {
 			put_wline(&s->as.text[2], 45., 0);
@@ -1333,7 +1345,8 @@ void put_history(void)
 	font = 0;
 	for (s = info['I' - 'A']; s != 0; s = s->next) {
 		u = s->as.text[0] - 'A';
-		if ((s2 = info[u]) == 0)
+		if (!(cfmt.fields[0] & (1 << u))
+		 || (s2 = info[u]) == 0)
 			continue;
 		if (!font) {
 			bskip(cfmt.textspace);
@@ -1379,6 +1392,7 @@ static char buf[STRL1];
 	b = buf;
 	r = &info['X' - 'A']->as.text[2];
 	if (first
+	 && (cfmt.fields[0] & (1 << ('X' - 'A')))
 	 && *r != '\0') {
 		if (strlen(p) + strlen(r) + 3 >= sizeof buf) {
 			error(1, 0, "Title or X: too long");
